local addonName, addon = ...

-- functions
local ipairs, pairs, type, unpack, wipe = ipairs, pairs, type, unpack, wipe
local format = string.format
local tconcat, tinsert, tsort = table.concat, table.insert, table.sort
local ChatEdit_GetActiveWindow = ChatEdit_GetActiveWindow
local ChatEdit_InsertLink = ChatEdit_InsertLink
local ChatFrame_OpenChat = ChatFrame_OpenChat
local CreateFrame = CreateFrame
local GetAutoCompleteRealms = GetAutoCompleteRealms
local GetMessageTypeColor = GetMessageTypeColor
local IsInGroup = IsInGroup
local IsInGuild = IsInGuild
local IsShiftKeyDown = IsShiftKeyDown
local SendChatMessage = SendChatMessage
-- constants
local DEFAULT_CHAT_FRAME = DEFAULT_CHAT_FRAME
local LE_PARTY_CATEGORY_INSTANCE = LE_PARTY_CATEGORY_INSTANCE
-- frames
local UIParent = UIParent
-- libs
local LibStub = LibStub
local Lib_EasyMenu = Lib_EasyMenu
local LibDataBroker = LibStub:GetLibrary("LibDataBroker-1.1")
local LibDBIcon = LibStub:GetLibrary("LibDBIcon-1.0")



local private = {}

private.menuFrame = CreateFrame("frame", "BrokerKeystonesContextMenuFrame", UIParent, "UIDropDownMenuTemplate")
private.menu = {
    Action = function(self, label, func)
        return {text = label, notCheckable = true, func = func}
    end,
    Checkbox = function(self, label, checked, func)
        return {text = label, keepShownOnClick = true, isNotRadio = true, checked = checked, func = func}
    end,
    Separator = function(self)
        return {text = " ", disabled = true, notCheckable = true}
    end,
    SubMenu = function(self, title, menuList)
        return {text = title, notCheckable = true, hasArrow = true, menuList = menuList}
    end,
    Title = function(self, title)
        return {text = title, isTitle = true, notCheckable = true}
    end,
    OptionToggle = function(self, label, key)
        return self:Checkbox(
            label,
            function() return addon.options[key] end,
            function() addon.options[key] = not addon.options[key] end
        )
    end,
}

function private:ContextMenu(menu)
    Lib_EasyMenu(menu, self.menuFrame, "cursor", 0, 0, "MENU")
end

private.tooltipLines = {}
function private:GetTooltipLines()
    wipe(private.tooltipLines)

    local charactersWithKeystones = addon:GetCharacters(function(_, c) return c.keystone ~= nil end)
    local numKeystones = addon:TableLength(charactersWithKeystones)

    local charactersWithRecords = addon.options.showRecords
        and addon:GetCharacters(function(_, c) return c.record ~= nil end)
        or {}
    local numRecords = addon:TableLength(charactersWithRecords)

    if numKeystones < 1 and numRecords < 1 then
        tinsert(private.tooltipLines, {"No Keystones", nil, 1, 1, 1})
        return private.tooltipLines
    end

    tinsert(private.tooltipLines, {
        "Character",
        format("%d |cffFFFFFF%s|r", numKeystones, numKeystones == 1 and "Keystone" or "Keystones"),
        1, 1, 1,
    })
    if not addon.options.showRecords then tinsert(private.tooltipLines, " ") end

    -- merge the two tables
    local characters = {}
    if numKeystones > 0 then
        for k, v in pairs(charactersWithKeystones) do
            characters[k] = v
        end
    end
    if numRecords > 0 then
        for k, v in pairs(charactersWithRecords) do
            characters[k] = v
        end
    end

    for key, character in pairs(characters) do
        if addon.options.showRecords then tinsert(private.tooltipLines, " ") end
        tinsert(private.tooltipLines, {
            addon:GetCharacterDisplayName(key, true),
            character.keystone
                and addon:GetKeystoneDisplayText(character.keystone, true)
                or format("|cff666666%s|r", "(no keystone)"),
            1.0, 1.0, 1.0,
            1.0, 1.0, 1.0,
        })
        if addon.options.showRecords and character.record then
            local mapId, level = unpack(character.record)
            -- todo: shorthand
            local instanceName = addon:GetMapName(mapId)
            tinsert(private.tooltipLines, {
                " ",
                format("Best: %s %d", instanceName, level),
                nil, nil, nil,
                0.7, 0.7, 0.7,
            })
        end
    end

    return private.tooltipLines
end

private.realms = {}
function private:GetCurrentRealms()
    wipe(private.realms)

    for _, realm in ipairs(GetAutoCompleteRealms()) do
        private.realms[realm] = true
    end

    private.realms[addon:GetRealmName()] = true

    return private.realms
end

function private:GetEligibleCharactersIterator(filter)
    local realms = self:GetCurrentRealms()

    return addon:GetCharactersIterator(function(_, c)
        return filter(_, c)
            and (addon.options.showOtherRealms == true or realms[c.realm])
            and (addon.options.showOtherFactions == true or c.faction == addon.player.faction)
    end)
end

private.keystones = {}
function private:BuildKeystoneTable(callback)
    wipe(self.keystones)
    for key, character in self:GetEligibleCharactersIterator(function(_, c) return c.keystone ~= nil end) do
        tinsert(self.keystones, format("%s (%s)",
            addon:GetKeystoneDisplayText(character.keystone),
            addon.options.anonymiseCharacters
                and addon:GetCharacterAnonymousDisplayName(key)
                or addon:GetCharacterDisplayName(key)
        ))
    end

    if #self.keystones > 0 then
        tsort(self.keystones)
        return callback(nil, self.keystones)
    end

    return callback("No keystones to report.")
end

function private:BuildKeystoneList(callback)
    return self:BuildKeystoneTable(function(err, tbl)
        if err then return callback(err) end

        return callback(nil, tconcat(tbl, ", "))
    end)
end

private.records = {}
function private:BuildRecordTable(callback)
    wipe(self.records)
    for key, character in self:GetEligibleCharactersIterator(function(_, c) return c.record ~= nil end) do
        local mapId, level = unpack(character.record)
        tinsert(self.records, format("%s %d (%s)",
            addon:GetMapName(mapId),
            level,
            addon.options.anonymiseCharacters
                and addon:GetCharacterAnonymousDisplayName(key)
                or addon:GetCharacterDisplayName(key)
        ))
    end

    if #private.records > 0 then
        tsort(self.records)
        return callback(nil, self.records)
    end

    return callback("No records to report.")
end

function private:BuildRecordList(callback)
    return self:BuildRecordTable(function(err, tbl)
        if err then return callback(err) end

        return callback(nil, tconcat(tbl, ", "))
    end)
end

function private:AddLinkToChat(link)
    local chatWindow = ChatEdit_GetActiveWindow()
    if chatWindow then
        ChatEdit_InsertLink(link)
    else
        ChatFrame_OpenChat(link, DEFAULT_CHAT_FRAME)
    end
end

function private:AddTextToChat(text)
    local chatWindow = ChatEdit_GetActiveWindow()
    if chatWindow then
        chatWindow:Insert(text)
    else
        ChatFrame_OpenChat(text, DEFAULT_CHAT_FRAME)
    end
end

function private:OnLeftClick()
    if IsShiftKeyDown() then
        if addon.player.keystone == nil then
            addon:Print("You don't have a keystone on this character.")
            addon:Print("If you did, shift-clicking the broker would link it to your chat.")

            return
        end

        return self:AddLinkToChat(addon.player.keystone)
    end

    local guildR, guildG, guildB = GetMessageTypeColor("GUILD")
    local guildHex = format("%x%x%x", guildR * 255, guildG * 255, guildB * 255)
    local partyR, partyG, partyB = GetMessageTypeColor("PARTY")
    local partyHex = format("%x%x%x", partyR * 255, partyG * 255, partyB * 255)

    local menu = self.menu
    self:ContextMenu({
        menu:Title("Report Keystones"),
        menu:Action("in Chat Box", function()
            self:BuildKeystoneList(function(err, list)
                if err then addon:Print(err) return end

                self:AddTextToChat(format("Keystones: %s", list))
            end)
        end),
        menu:Action(format("|cff%s%s|r", guildHex, "to Guild"), function()
            self:BuildKeystoneList(function(err, list)
                if err then addon:Print(err) return end

                if IsInGuild() then
                    SendChatMessage(format("Keystones: %s", list), "GUILD")
                else
                    addon:Print("You are not in a guild.")
                end
            end)
        end),
        menu:Action(format("|cff%s%s|r", partyHex, "to Group"), function()
            self:BuildKeystoneList(function(err, list)
                if err then addon:Print(err) return end

                if IsInGroup(LE_PARTY_CATEGORY_INSTANCE) and not IsShiftKeyDown() then
                    SendChatMessage(format("Keystones: %s", list), "INSTANCE_CHAT")
                elseif IsInGroup() then
                    SendChatMessage(format("Keystones: %s", list), "PARTY")
                else
                    addon:Print("You are not in a party or instance.")
                end
            end)
        end),
        menu:Separator(),
        menu:Title("Report Weekly Records"),
        menu:Action("in Chat Box", function()
            self:BuildRecordList(function(err, list)
                if err then addon:Print(err) return end

                self:AddTextToChat(format("Weekly Records: %s", list))
            end)
        end),
        menu:Action(format("|cff%s%s|r", guildHex, "to Guild"), function()
            self:BuildRecordList(function(err, list)
                if err then addon:Print(err) return end

                if IsInGuild() then
                    SendChatMessage(format("Weekly Records: %s", list), "GUILD")
                else
                    addon:Print("You are not in a guild.")
                end
            end)
        end),
        menu:Action(format("|cff%s%s|r", partyHex, "to Group"), function()
            self:BuildRecordList(function(err, list)
                if err then addon:Print(err) return end

                if IsInGroup(LE_PARTY_CATEGORY_INSTANCE) and not IsShiftKeyDown() then
                    SendChatMessage(format("Weekly Records: %s", list), "INSTANCE_CHAT")
                elseif IsInGroup() then
                    SendChatMessage(format("Weekly Records: %s", list), "PARTY")
                else
                    addon:Print("You are not in a party or instance.")
                end
            end)
        end),
    })
end

function private:OnRightClick()
    local menu = self.menu
    self:ContextMenu({
        menu:Title(addonName),
        menu:SubMenu("Options", {
            menu:Title("General"),
            menu:OptionToggle(
                "Use shorthand dungeon names where available",
                "useShorthand"
            ),
            menu:OptionToggle(
                "Show characters' weekly records in tooltip",
                "showRecords"
            ),
            menu:Separator(),
            menu:Title("Data Reporting"),
            menu:OptionToggle(
                "Use characters' class names instead of their real names",
                "anonymiseCharacters"
            ),
            menu:OptionToggle(
                "Include characters from other realms",
                "showOtherRealms"
            ),
            menu:OptionToggle(
                "Include characters from other factions",
                "showOtherFactions"
            ),
        }),
        menu:SubMenu("Miscellaneous", {
            menu:OptionToggle(
                "Debug mode",
                "debug"
            ),
            menu:Separator(),
            menu:Action("Delete Keystone Data (no confirmation)", function()
                addon:DeleteAllCharacterKeystones()
            end),
            menu:Action("Delete Weekly Record Data (no confirmation)", function()
                addon:DeleteAllCharacterRecords()
            end),
        }),
    })
end



local broker = LibDataBroker:NewDataObject(addonName, {
    type = "data source",
    label = "Keystone",
    text = "No Keystone",
    icon = [[Interface\Icons\inv_relics_hourglass]],
})

LibDBIcon:Register(addonName, broker, {})

function broker:OnTooltipShow()
    local lines = private:GetTooltipLines()

    for _, line in pairs(lines) do
        if type(line) == "string" then
            line = {line}
        end

        self:AddDoubleLine(unpack(line))
    end
end

function broker:OnClick(btn)
    if btn == "LeftButton" then return private:OnLeftClick() end
    if btn == "RightButton" then return private:OnRightClick() end
end

addon:RegisterCallback(addon.EVENT_KEYSTONE, function(_, key, link)
    broker.text = link and addon:GetKeystoneDisplayText(link, true) or "No Keystone"
end)

addon:RegisterCallback(addon.EVENT_OPTION_CHANGED, function(_, key)
    if key ~= "useShorthand" then return end

    local character = addon:GetCharacter()
    if not character or not character.keystone then
        broker.text = "No Keystone"
        return
    end

    broker.text = addon:GetKeystoneDisplayText(character.keystone, true)
end)

addon:RegisterCallback(addon.EVENT_OPTION_CHANGED, function(_, key, value)
    if key ~= "showMinimapButton" then return end

    if value then
        LibDBIcon:Show(addonName)
    else
        LibDBIcon:Hide(addonName)
    end
end)
